module Observe
  ( ObserveResult(..)
  ,   isFound
  ,   isNotFound
  ,   isInterrupted
  ,   fromFound
  , newObservation	-- :: String -> Maybe String -> Bool -> ObserveResult
  , newObservationSrc	-- :: String -> Int -> Int -> ObserveResult
  , observableInfo	-- :: IO ()
  ) where

import LowLevel
import FFIExtensions

------------------------------------------------------------------------
data ObserveResult = Found [FileNode]
                   | NotFound
                   | Interrupted

isInterrupted :: ObserveResult -> Bool
isInterrupted Interrupted = True
isInterrupted _           = False

isNotFound :: ObserveResult -> Bool
isNotFound NotFound       = True
isNotFound _              = False

isFound :: ObserveResult -> Bool
isFound (Found _)         = True
isFound _                 = False

fromFound :: ObserveResult -> [FileNode]
fromFound (Found r)         = r
fromFound _                 = []

------------------------------------------------------------------------
foreign import setObserveContext :: Bool -> CString -> IO ()
foreign import lookForFirstApp   :: Bool -> CString -> IO FileNode
foreign import lookForFirstSrc   :: Int -> Int -> CString -> IO FileNode
foreign import nextObservation   :: FileNode -> IO FileNode
foreign import observableInfo    :: IO ()

-- Get all observations for ident1 (within ident2 if available)
-- If ident2 is Nothing, then the `recursive' flag selects whether
-- calls from ident1 to itself are included.
newObservation :: String -> Maybe String -> Bool -> ObserveResult
newObservation ident1 ident2 rec =
  unsafePerformIO $
  do _ <- case ident2 of
              Nothing -> withCString "" (setObserveContext False)
              Just id -> withCString id (setObserveContext True)
     n <- withCString ident1 (lookForFirstApp rec)
     case int n of
       1 -> return NotFound
       3 -> return Interrupted
       _ -> return (Found (n: moreObservations n))

moreObservations :: FileNode -> [FileNode]
moreObservations n =
  unsafePerformIO $
  do n' <- nextObservation n
     case int n' of
       1 -> return []
       3 -> return []
       _ -> return (n': moreObservations n')

-- Get all observations at the given source location.
newObservationSrc :: String -> Int -> Int -> ObserveResult
newObservationSrc moduleName line column =
  unsafePerformIO $
  do withCString "" (setObserveContext False)
     n <- withCString moduleName (lookForFirstSrc line column)
     case int n of
       1 -> return NotFound
       3 -> return Interrupted
       _ -> return (Found (n: moreObservations n))

------------------------------------------------------------------------
