Return-Path: colin@cs.york.ac.uk
Return-path: <colin@cs.york.ac.uk>
Envelope-to: malcolm@cs.york.ac.uk
Received: from pc179 ([144.32.41.180] ident=colin)
	by minster.cs.york.ac.uk with smtp (Exim 3.22 #6)
	id 16nzcz-0001h3-00
	for malcolm@cs.york.ac.uk; Thu, 21 Mar 2002 10:17:01 +0000
Received: by pc179 (sSMTP sendmail emulation); Thu, 21 Mar 2002 10:17:01 +0000
From: Colin Runciman <colin@cs.york.ac.uk>
Date: Thu, 21 Mar 2002 10:17:01 +0000
To: malcolm@cs.york.ac.uk
Message-Id: <E16nzcz-0001h3-00@minster.cs.york.ac.uk>

-- Ordering of partially computed graphs, for hat-observe.
-- Version 1: comparison of subgraphs in a single graph;
-- node structure closer to .hat format.
-- Colin Runciman, March 2002

import FiniteMap

data SLMD = Same | Less | More | Diff deriving Show

lub Diff _    = Diff
lub More Diff = Diff
lub More Less = Diff
lub More More = More
lub More Same = More
lub Less Diff = Diff
lub Less More = Diff
lub Less Less = Less
lub Less Same = Less
lub Same c    = c

data Node a = Uneval | Ident String | Applic [a]

type GraphRef = Int
type Graph    = [Node GraphRef]

cmpGraph :: GraphRef -> GraphRef -> SLMD
cmpGraph g1 g2 = fst (cmpGraphInCxt g1 g2 (Same,emptyFM))

type Context = (SLMD, FiniteMap (GraphRef,GraphRef) SLMD)

cmpGraphInCxt :: GraphRef -> GraphRef -> Context -> Context
cmpGraphInCxt _  _  (Diff,m) = (Diff,m)
cmpGraphInCxt r0 r1 (c,m)    =
  if r0 == r1 then (c,m)
  else
    case applyFM m (r0,r1) of
    Nothing ->
      let (c',m) = cmpNode (graph !! r0) (graph !! r1) in
      (c', overFM m (r0,r1) c')
    Just c' ->
      (c `lub` c',m)
  where
  cmpNode Uneval      Uneval      = (c,m)
  cmpNode Uneval      _           = (c `lub` Less,m)
  cmpNode _           Uneval      = (c `lub` More,m)
  cmpNode (Ident s0)  (Ident s1)
                      | s0 == s1  = (c,m)
  cmpNode (Applic xs) (Applic ys)
         | length xs == length ys = foldr
                                      (uncurry cmpGraphInCxt)
                                      (c, overFM m (r0,r1) Same)
                                      (zip xs ys)
  cmpNode _            _          = (Diff,m)
 
---------------8<-------Just for testing----8<---------------
  
graph :: Graph
graph = [ Applic [1, 2, 3]   -- node 0
	, Ident ":"          -- node 1
	, Ident "a"          -- node 2
	, Applic [1, 2, 0]   -- node 3
	, Uneval             -- node 4
	]

main = print (cmpGraph 0 3)
